def train_router():
    raise NotImplementedError


def delete_router():
    raise NotImplementedError


def rename_router():
    raise NotImplementedError


def list_routers():
    raise NotImplementedError
