def create_router_config():
    raise NotImplementedError


def get_router_config():
    raise NotImplementedError


def delete_router_config():
    raise NotImplementedError


def rename_router_config():
    raise NotImplementedError


def list_router_configs():
    raise NotImplementedError
